;----------------------------------------------------------------
;	Title	: callgate.asm
;	Date	: 3-5-97
;	Author	: Mark Willis and others   73150.730@compuserve.com 
;	Descr	: Sample program for using a 386 callgate to jump
;			  from ring3 code to ring0 code and back 
;
;		This program:
;		a) Converts various physical addresses to linear
;		b) Switches to infected mode privledge level 0
;		c) Does a 48bit ret far to a ring 3 segment
;		d) Call thru a gate back to ring 0
;		e) Does a 48bit ret far back to ring 3 segment
;
;	It took four months to get this working :(
;----------------------------------------------------------------
.386p

;----------------------------------------------------------------
; Structure for descriptor
;----------------------------------------------------------------
dscp struc
		limit1		dw	0
        base1		dw	0
        base2		db	0
        des_type	db	0
        limit2		db	0
        base3		db	0
dscp	ends

;----------------------------------------------------------------
; Macros
;----------------------------------------------------------------
callf	macro selector
	db 09ah
	dw 0
	dw selector
endm	

;----------------------------------------------------------------
;Global Descriptor Table
;
; video mem starts at
; phys	:    B800 0000
; linear: 0B 8000 0000
;----------------------------------------------------------------
des_code0	equ	08h			;this must be an equ ???
des_video	equ	10h 		;descr for vid mem is at offset 10h of GDT
des_code3	equ	1bh 		;stack selector ring 3 <-watch extra bits !!!
des_ss0		equ	20h
des_tss		equ	28h
des_data3	equ	33h			;40h + 3h (priv level 3) = 43h
des_data0	equ	38h
des_stack3	equ	43h
des_gate	equ 4bh			;second ring 0 selector

GDT segment para public use16 'GDT'
gdt_tab		label qword
	dscp	<,,,,,>						; 00h first must be null or so they say 
    dscp    <0ffffh,,,09ah,,>   		; 08h code level 0
	dscp	<0ffffh,8000h,0bh,0f2h,,>	; 10h video descr permissions=f2
    dscp    <0ffffh,,,0fah,,>			; 18h code  level 3
    dscp    <stack0_top,,,92h,,>		; 20h stack level 0
	dscp	<tss_limit,,,089h,,>		; 28h tss descriptor
	dscp	<gdata_limit,,,0f2h,,>		; 30h ring 3 data descriptor
	dscp	<gdata_limit,,,0f2h,,>		; 38h ring 0 data descriptor
	dscp	<stack3_top,,,0f2h,,>		; 40h ring 3 stack descriptor
    dscp    <backto_ring0,des_code0,,0ech,,>   		; 48h call gate
gdt_limit	equ $-gdt_tab	
GDT ends

;----------------------------------------------------------------
; Data segment
;----------------------------------------------------------------
Gdata segment para public use16 'Gdata'
pGDT	label fword 		;This will be loaded into GDT register 
pGDT_limit	dw ?			;GDT limit
pGDT_addr	dd ?			;GDT base address


screen_pos	dw	320			;upper left corner of screen
ring3_msg	db '<Look ma, ring 3>',0			
ring3_msg2	db '<ring 3 again>',0
ring0_msg	db ' <Ring 0> ',0
ring0_msg2	db ' <Thru the callgate to ring 0 !> ',0
video_addr	dw	des_video
			db 100 dup(0)	;stack space
Gdata_limit		equ $		;use the current pos to meas the size of seg
Gdata 			ends

;----------------------------------------------------------------
; Ring 3 stack
;----------------------------------------------------------------
ring3_stack 	segment para public use16 'ring3_stack'
				db 'ring3 stack     '
				db 100h dup(0)
stack3_top	equ $
ring3_stack		ends
	
;----------------------------------------------------------------
; Ring 0 stack
;----------------------------------------------------------------
stack0_seg segment para public use16 'stack0_seg'
	db	100h dup(0)
stack0_top equ $
stack0_seg ends	


;----------------------------------------------------------------
; TSS segment 
;----------------------------------------------------------------
tss_seg	segment para public use16 'tss_seg'
dd	0						; Previous TSS
dd  offset stack0_top		;stack level 0 ptr
dd	des_ss0					;stack level 0 seg
dd	0						;stack level 1 ptr
dd	0						;stack level 1 segment
dd	0						;stack level 2 ptr
dd	0						;stack level 2 segment

dd	0						; the cr3 register
dd	offset	0				; eip
dd	0						; flag
dd	0						; eax
dd	0						; ecx
dd	0						; edx
dd	0						; ebx
dd	offset	0h				; esp
dd	0						; ebp
dd	0						; esi
dd	0						; edi

dd	des_data3	; es
dd	des_code3	; cs
dd	des_data3	; ss
dd	des_data3	; ds
dd 	des_data3	; fs
dd	des_data3	; gs
dd	0			; LDT
dw	0			; task trap flag
dw	68h			; I/O base
tss_limit	equ $
tss_seg ends

;----------------------------------------------------------------
; Code segment
;----------------------------------------------------------------
code 	segment para public use16 'code'
		assume  cs:code, ds:gdata
mainb	proc far
		
        mov ax,gdata				; get data seg where GDT ptr lives
        mov ds,ax					; put in DS
        mov ss,ax
        mov sp,offset Gdata_limit
		mov gs,ax
		mov es,ax
		
		
		call get_lin_gdt 			;Calc the lin addr of GDT and store 
		call store_lin_cs			;Store the linear address of the code 
		call fix_code3				;Store lin addr of 2nd code seg	
		call fix_tss
		call fix_stacks
		call fix_data

									;Flight attendants, prepare for takeoff
		cli							;hold all interrupts
		lgdt [pGDT]					;load GDT addr and limit into GDTR
		mov eax,cr0					;get cr0 register
		or al,1						;set protected mode enable
		mov cr0,eax					;ka-boom! we're in protect mode 
		jmp dword ptr cs:[enter_prot]; far jump to flush instruct queue
		
enter_prot:
		dw offset pmode_begin		;EIP
		dw des_code0				;code segment selector
		
;*************************************************************************		
pmode_begin:					;NOW IN PROTECTED MODE !!

	xor ax,ax					;not using LDTs here
	lldt ax						;so put 0 into ldt register
	
	mov ax,des_tss				;going thru the callgate will cause the tss
	ltr ax						;to be looked at to get the temp ring0 stack	
								;so we gotta setup a tss	
								
	mov es,video_addr			;put a message on the screen
	mov si,offset ring0_msg
	call show_string
	
	mov ax,des_data3			;Prepare for jump to ring 3				
	mov ds,ax					;The retf won't set up these registers,so let's
	mov es,ax					;set them up with a ring3 data selector
	mov fs,ax
	mov gs,ax
	
	xor eax,eax
	mov ax,des_stack3
	push eax					;push stack seg
	
	xor eax,eax
	mov ax,offset stack3_top	;push esp
	push eax
	
	xor eax,eax
	mov ax,des_code3			;push cs
	push eax
	
	mov ax, offset code3_begin
	push eax					;push eip
	
	db 66h
	retf						;wala - a far call to a ring3 segment !!!


;------------------------------------------------------------------------
backto_ring0:					;Here's where we land when we go	
	mov ax,des_data0			;thru the callgate
	mov ds,ax

	mov si,offset ring0_msg2	;display back to ring 0 msg
	call show_string
	db 66h
	retf
	

;****************************************************************		
; End of main code
;****************************************************************		



;----------------------------------------------------------------
; store_lin_cs
;----------------------------------------------------------------
store_lin_cs:
		mov ax,GDT					;setup esdi for write 
		mov es,ax					;into gdt code sel
		mov di,des_code0
		
		push cs						;ax=seg bx=ofs  
		pop ax
		mov bx,0
		call phys_to_lin			;convert to linear
		call store_addr				;store in gdt
		retn
		
;------------------------------------------------------------------------
; get_lin_gdt
;------------------------------------------------------------------------
get_lin_gdt:
		mov	ax,offset gdt_limit		;get GDT segment limit (size)
        mov	pGDT_limit,ax			;put in pGDT_limit
        xor	eax,eax					;zero out eax
        mov ax,GDT					;get GDT segment address
        shl eax,4					;convert to 32 bit linear address
        mov pGDT_addr,eax			;put in pGDT address
        retn
		
				
;------------------------------------------------------------------------
; show_char  al = char
;------------------------------------------------------------------------
show_char:
		push es
		push di
		push ax
		mov ax,des_video		;load segment (descriptor) of dest
		mov es,ax
		mov ax,screen_pos		;offs 320 is top left of screen
		mov di,ax
		pop ax
		mov ah,07h				;normal video attribute
		mov es:[di],ax			;write it onto the screen
		mov ax, screen_pos
		add ax,2
		mov screen_pos,ax	
		pop di
		pop es
		retn
		
				
;------------------------------------------------------------------------
; show_string  IN:si=string offset
;------------------------------------------------------------------------
show_string:
		push ax
show_loop:		
		xor ax,ax
		mov al,[si]
		or al,al
		jz show_string_end
		call show_char	
		inc si		
		jmp show_loop
show_string_end:		
		pop ax
		retn
		
		
;------------------------------------------------------------------------
; phys_to_lin   IN:ax=segment  bx=offset OUT: ax:bl= 32bit linear (msb:lsb)
;------------------------------------------------------------------------
phys_to_lin:
		push cx
		push ax
		shl ax,4
		add ax,bx
		mov cx,ax
		pop ax 
		shr ax,12
		mov al,bl
		mov ax,cx
		pop cx
		retn
		
		
;------------------------------------------------------------------------
; store_addr IN:ax=linaddr (lsw)  bl=msb es=gdt  di=descriptor
;
; A selector of 08 means 8 bytes into the GDT. Simple enough ?  But if
; that selector were ring 3 its value would be 0bh (8 + 3).  So when
; using the selector val as an index into the GDT, first we gotta mask
; off bits 0 and 1 (privledge level) and bit 2 (GDT or LDT flag)
;------------------------------------------------------------------------
store_addr:
	push ax
	mov ax,di			;mask off rpl and ti bits
	and ax,0f8h			;this will be our offset into the GDT
	mov di,ax
	pop ax
	
	push ax				;the GDT segment
	mov ax,GDT
	mov es,ax
	pop ax
	
	
	add di,2			;move past bytes 0 and 1 of the descriptor (limit)
	mov es:[di],ax		;write the linear address
	add di,2
	mov al,bl
	mov es:[di],al
	
skipit:	
	retn
	
;------------------------------------------------------------------------
; fix_code3 - fix the descriptor for a ring 3 code segment 
;------------------------------------------------------------------------
fix_code3:
	mov ax,GDT
	mov es,ax
	mov di,des_code3
	
	mov ax,seg2
	mov bx,0
	call phys_to_lin
	call store_addr
	retn
	

	
;------------------------------------------------------------------------
; fix_tss - fix the tss descriptor
;------------------------------------------------------------------------
fix_tss:
	mov ax,GDT
	mov es,ax
	mov di,des_tss
	
	mov ax,tss_seg
	mov bx,0
	call phys_to_lin
	call store_addr
	retn
	
	
;------------------------------------------------------------------------
; fix_stacks - fix the stack selectors
;------------------------------------------------------------------------
fix_stacks:
	mov ax,GDT			;fix ring 0 stack
	mov es,ax
	mov di,des_ss0
	
	mov ax,stack0_seg	;stack segment
	mov bx,0			;offset 0 into segment
	call phys_to_lin	;convert to 32 bit linear
	call store_addr		;shove lin address into GDT
	
	
	
		
	mov ax,GDT			;fix ring 3 stack
	mov es,ax
	mov di,des_stack3
	
	mov ax,ring3_stack
	mov bx,0
	call phys_to_lin
	call store_addr
	retn	
	
	
;------------------------------------------------------------------------
; fix_data - fill in the address of the data segment.  We will use
; 			ring 0 and ring3 selector to this memory
;------------------------------------------------------------------------
fix_data:
	mov ax,GDT			;convert the ring 3 selector to linear
	mov es,ax
	mov di,des_data3
	
	mov ax,Gdata
	mov bx,0
	call phys_to_lin
	call store_addr
	
	
	mov ax,GDT			;convert the ring 0 selector
	mov es,ax
	mov di,des_data0
	
	mov ax,Gdata
	mov bx,0
	call phys_to_lin
	
	call store_addr
	
	retn

mainb 	endp
code	ends

;********************************************************************
; seg2 - this is where the far call lands
;********************************************************************
seg2 	segment para public use16 'seg2'
		assume  cs:seg2, ds:gdata
		
newseg	proc far		

code3_begin:		
		mov si,offset ring3_msg
		call show_string3
		callf des_gate
		
		mov si,offset ring3_msg2
		call show_string3
		
loop_forever:					;loop here forever to keep pc
		jmp loop_forever		;from rebooting
		
;******** PROGRAM ENDS HERE *******			
			
;------------------------------------------------------------------------
; show_char  al = char
;------------------------------------------------------------------------
show_char3:
		push es
		push di
		push ax
		mov ax,des_video		;load segment (descriptor) of dest
		mov es,ax
		mov ax,screen_pos		;offs 320 is top left of screen
		mov di,ax
		pop ax
		mov ah,07h				;normal video attribute
		mov es:[di],ax			;write it onto the screen
		mov ax, screen_pos
		add ax,2
		mov screen_pos,ax	
		pop di
		pop es
		retn
		
				
;------------------------------------------------------------------------
; show_string  IN:si=string offset
;------------------------------------------------------------------------
show_string3:
		push ax
show_loop3:		
		xor ax,ax
		mov al,[si]
		or al,al
		jz show_string_end3
		call show_char3	
		inc si		
		jmp show_loop3
show_string_end3:		
		pop ax
		retn
		
newseg	endp		
seg2	ends		
end 	mainb

